'use strict';

Object.defineProperty(exports, '__esModule', {
  value: true
});
var map = require('map-ast');

var transforms = {
  JSXElement: function JSXElement(_ref, env) {
    var _ref$openingElement = _ref.openingElement;
    var name = _ref$openingElement.name;
    var attributes = _ref$openingElement.attributes;
    var children = _ref.children;

    var expr = {
      type: 'CallExpression',
      callee: { type: 'Identifier', name: 'JSX' },
      arguments: [parseCallee(name, env)]
    };
    if (attributes.length) addArg(1, expr, parseAttrs(attributes, env));
    if (children.length) addArg(2, expr, parseChildren(children, env));
    return expr;
  },
  JSXExpressionContainer: function JSXExpressionContainer(node, env) {
    return map(transforms, env, node.expression);
  },
  JSXText: function JSXText(node) {
    return literal(node.value);
  }
};

var addArg = function addArg(index, call, arg) {
  while (call.arguments.length < index) {
    call.arguments.push(literal(null));
  }
  call.arguments.push(arg);
};

var parseAttrs = function parseAttrs(attributes, env) {
  var out = call('assign');
  var attrs = { type: 'ObjectExpression', properties: [] };

  var _iteratorNormalCompletion = true;
  var _didIteratorError = false;
  var _iteratorError = undefined;

  try {
    for (var _iterator = attributes[Symbol.iterator](), _step; !(_iteratorNormalCompletion = (_step = _iterator.next()).done); _iteratorNormalCompletion = true) {
      var attr = _step.value;

      if (attr.type == 'JSXSpreadAttribute') {
        if (attrs.properties.length) {
          out.arguments.push(attrs);
          attrs = { type: 'ObjectExpression', properties: [] };
        }
        out.arguments.push(attr.argument);
        continue;
      }
      var name = attr.name;
      var value = attr.value;

      attrs.properties.push({
        type: 'ObjectProperty',
        kind: 'init',
        key: toKey(name),
        value: value == null ? { type: 'Identifier', name: name.name } : map(transforms, env, value)
      });
    }
  } catch (err) {
    _didIteratorError = true;
    _iteratorError = err;
  } finally {
    try {
      if (!_iteratorNormalCompletion && _iterator['return']) {
        _iterator['return']();
      }
    } finally {
      if (_didIteratorError) {
        throw _iteratorError;
      }
    }
  }

  if (attrs.properties.length) out.arguments.push(attrs);

  // was Object.assign necessary
  if (out.arguments.length == 1) out = out.arguments[0];

  // is the first attribute a spread
  else if (out.arguments[0] === attributes[0].argument) {
      // prevent mutation of potentially shared object
      out.arguments[0] = call('create', out.arguments[0]);
    }

  return out;
};

var toKey = function toKey(ident) {
  return (/^[a-zA-Z$][\w$]*$/.test(ident.name) ? reuse(ident) : literal(ident.name)
  );
};

var call = function call(property, arg) {
  return {
    type: 'CallExpression',
    callee: {
      type: 'MemberExpression',
      object: { type: 'Identifier', name: 'Object' },
      property: { type: 'Identifier', name: property }
    },
    arguments: arg ? [arg] : []
  };
};

var parseChildren = function parseChildren(children, env) {
  // remove whitespace between nodes
  children = children.filter(function (child) {
    return child.type != 'JSXText' || !child.value.match(/^[ \t]*[\r\n][ \t\r\n]*$/);
  });
  // remove leading and trailing new lines
  var end = children.length - 1;
  children[0] = trim('left', children[0]);
  children[end] = trim('right', children[end]);
  return {
    type: 'ArrayExpression',
    elements: children.map(function (child) {
      return map(transforms, env, child);
    })
  };
};

var trim = function trim(side, child) {
  if (child.type == 'JSXText') child.value = child.value.replace(trim[side], '');
  return child;
};

trim.left = /^(?:[\r\n]+\s*)+/;
trim.right = /(?:\s*[\r\n]+)+$/;

var parseCallee = function parseCallee(node, env) {
  return getTarget(node) in env ? reuse(node) : literal(toString(node));
};

var getTarget = function getTarget(_x) {
  var _again = true;

  _function: while (_again) {
    var node = _x;
    _again = false;

    if (node.type == 'JSXMemberExpression') {
      _x = node.object;
      _again = true;
      continue _function;
    } else {
      return node.name;
    }
  }
};var reuse = function reuse(node) {
  if (node.type == 'JSXMemberExpression') {
    node.type = 'MemberExpression';
    reuse(node.object);
    reuse(node.property);
  } else {
    node.type = 'Identifier';
  }
  return node;
};

var toString = function toString(node) {
  return node.type == 'JSXMemberExpression' ? toString(node.object) + '.' + toString(node.property) : node.name;
};

var literal = function literal(value) {
  var type;
  if (typeof value == 'number') type = 'NumberLiteral';else if (typeof value == 'string') type = 'StringLiteral';else if (typeof value == 'boolean') type = 'BooleanLiteral';else if (value === null) type = 'NullLiteral';else type = 'RegexLiteral';
  return { type: type, value: value };
};

var babel_plugin = function babel_plugin() {
  return {
    visitor: {
      Program: function Program(path) {
        path.node.body = map(transforms, null, path.node).body;
      }
    }
  };
};

var JSX = function JSX(ast) {
  return map(transforms, null, ast);
};

exports['default'] = JSX;
exports.babel_plugin = babel_plugin;
exports.JSX = JSX;

